/********************************************************\
* bfi89.c -- brainf*ck interpreter for TI-89             *
* by james longstreet 2002. protected under version 2 of *
* the GNU General Public License.                        *
* see http://www.gnu.org/licenses/gpl.html for details.  *
*                                                        *
* version 0.8
* see file history.txt for complete version history.     *
\********************************************************/
 
#define USE_TI89              // Compile for TI-89
#define USE_TI92PLUS          // Compile for TI-92 Plus
#define USE_V200              // Compile for V200
#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
#define SAVE_SCREEN           // Save/Restore LCD Contents
#include <tigcclib.h>         // Include All Header Files

#define ESC_ROW (0xFFBF)        // bf? coincidence...?
#define ESC_COL (0x0001)


int  p, r, k;
char *a = NULL, *f = NULL, b, o, *s;
FILE *in, *out;
char (*get_a_char)();
short (*put_a_char)(char ch);
void* kbq;

void free_all()
{
    if (a)    free(a);
    if (f)    free(f);
}

void die()
{
	printf("\ndone");
    ngetchx();
    free_all();
    exit(0);
}

/* functions to get/put bytes to/from screen, keyboard, file */
/* function pointers set to either getc_file or getc_kbd and */
/* putc_file or putc_scr, then get_a_char/put_a_char called */
char getc_file()
{
	char ch = fgetc(in);
    return (ch != EOF) ? ch : 0;
}

char getc_kbd()
{
	int k = ngetchx();
	if (k == KEY_ESC)
		die();
	if (k == KEY_ON)
		return 0;
	else	{
		putchar((char)k);
		return (char)k;
	}
}

short putc_file(char ch)
{
	return fputc(ch,out);
}

short putc_scr(char ch)
{
	return putchar(ch);
}

void check_esc()
{
	/*if (_rowread(ESC_ROW) == ESC_COL)
		die();
	GKeyFlush();*/
	unsigned short key;
	OSdequeue(&key,kbq);
	if (key == KEY_ESC)
		die();
}

/* do the actual interpreting. partially stolen from original */
/* bfi.c, but bfi.c sucks... there's a lot of things wrong with it */
/* imho. it won't compile as is on freebsd, and it doesn't really */
/* work anyway. so basically, the switch block was taken. */
void interpret(char *c)
{
		check_esc();
    char *d;
    r++;
    while( *c ) {
        //if(strchr("<>+-,.[]\n",*c))printf("%c",*c);
        check_esc();
        
        switch(o=1,*c++) {
            case '<': p--;        break;
            case '>': p++;        break;
            case '+': a[p]++;     break;
            case '-': a[p]--;     break;
            case '.': if (a[p])
            						 put_a_char(a[p]);
            						 /*fflush(stdout);*/ break;
                /* disgusting. if keypress is KEY_ON, put 0.
                otherwise, put the keypress and print */
            case ',':
                a[p] = get_a_char();
                break;
            case '[':
                for( b=1,d=c; b && *c; c++ )
                    b+=*c=='[', b-=*c==']';
                if(!b) {
                    c[-1]=0;
                    while( a[p] )
                        interpret(d);
                    c[-1]=']';
                    break;
                }
                case ']':
                    puts("UNBALANCED BRACKETS"), die();
                case '#':
                    /*printf("%2d %2d %2d %2d %2d %2d %2d %2d %2d %2d\n%*s\n",
                    *a,a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9],3*p+2,"^");*/
                    break;
                default: o=0;
        }
        if( p<0 || p>10000)
            puts("RANGE ERROR"), die();
    }
    r--;
    //	chkabort();
}
void _main()
{
    FILE *z;
    ESI argptr;
    char filename[10];

    clrscr();

    a = (char*)malloc(10000);
    f = (char*)malloc(10000);
    s=f;

    if (!a || !f)	{
        printf("needs 20kb free to run\n");
        die();
    }

    InitArgPtr(argptr);	/* set pointer to stack*/
    if (ArgCount() == 0)	{
        printf("no input file.\n");
        die();
    }
    strcpy(filename,GetStrnArg(argptr));
    z = fopen(filename,"r");
    if (z == NULL)	{
        printf("couldn't open %s\n",filename);
        die();
    }
    
    kbq = kbd_queue();

    if (RemainingArgCnt(argptr)) /* input file specified */	{
        strcpy(filename,GetStrnArg(argptr));
        if (strcmp(filename,"-") != 0)	/* - = stdin */	{
            in = fopen(filename,"r");
            if (in == NULL)	{
                printf("can't open in: %s",filename);
                die();
            }
            get_a_char = &getc_file;
        } else get_a_char = &getc_kbd;
    } else get_a_char = &getc_kbd;

    if (RemainingArgCnt(argptr)) /* output file specified */ {
        strcpy(filename,GetStrnArg(argptr));
        if (strcmp(filename,"-") != 0)	/* - = stdout */ {
            out = fopen(filename,"w");
            if (out == NULL)	{
                printf("can't open out: %s",filename);
                die();
            }
            put_a_char = &putc_file;
        } else put_a_char = &putc_scr;
    } else put_a_char = &putc_scr;

    p = 0;
    /* lousy tigcclib doesn't define bzero */
    memset(a,0,10000);
    memset(f,0,10000);

    while( (b=getc(z))>0 )
        *s++=b;
    *s=0;
    clrscr();

    interpret(f);

    die();
}
